from google.cloud import bigquery, storage
import pydata_google_auth

from functools import lru_cache


def google_client(service, project_id=None):
    return _clients(project_id)[service]


@lru_cache(256)
def _clients(project_id=None):

    return dict(
        bigquery=bigquery.Client(project=project_id, credentials=credentials()),
        storage=storage.Client(project=project_id, credentials=credentials()),
    )


def credentials():

    SCOPES = [
        "https://www.googleapis.com/auth/cloud-platform",
    ]

    return pydata_google_auth.get_user_credentials(
        SCOPES,
        # Use the NOOP cache to avoid writing credentials to disk.
        # credentials_cache=pydata_google_auth.cache.NOOP,
    )


from google.cloud import bigquery

from basedosdados.download.base import google_client


def freq(
    dataset_id,
    table_id,
    col_name,
    project_id="basedosdados",
    order_by_freq=False,
    ascending=True,
):
    """
    Return a dataframe containing the frequency table of the `col_name`
    (which can be a list of columns) specified.


    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        table_id (:obj:`str`): Table id available in project_id.dataset_id.
        col_name (:obj:`str`, list): Column available in project_id.dataset_id.table_id.
        project_id (:obj:`str`, optional):n case you want to use to
            query another project. Defaults to "basedosdados".
        order_by_freq (bool, optional): Order results by frequency.
            Defaults to False.
        ascending (bool, optional): Whether results should ascend. Defaults to True.

    Returns:
        [type]: [description]
    """

    client = google_client("bigquery", project_id=project_id)
    if not (isinstance(col_name, list) or isinstance(col_name, tuple)):
        col_name = [col_name]

    # generate the query for the frequency table
    table_name = f"{project_id}.{dataset_id}.{table_id}"
    query_str = _freq_query(table_name, col_name)

    freq_cols = ["perc_" + col.lower() for col in col_name]
    cols_str = " and ".join(col_name)

    if order_by_freq:
        job = client.query(query_str)
        result = job.to_dataframe().sort_values(by="frequency", ascending=ascending)
    else:
        job = client.query(query_str)
        result = job.to_dataframe().sort_values(by=col_name, ascending=ascending)

    return result


def _freq_query(table_name, col_name):
    keys = ", \n\t\t".join(col_name)

    # each calculated column of the frequency table
    counts = [
        "COUNT(*) AS frequency",
        "COUNT(*)/(SUM(COUNT(*)) OVER()) AS percentage",
        "SUM(COUNT(*)) OVER window_freq AS cumulative_freq",
        "(SUM(COUNT(*)) OVER window_freq)/(SUM(COUNT(*)) OVER ()) AS cumulative_percentage",
    ]

    aggregations = ",\n\t".join(counts)

    query_str = f"""
        SELECT {keys},
            {aggregations}
        FROM {table_name}
        GROUP BY {keys}
    WINDOW window_freq AS (ORDER BY COUNT(*) RANGE BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)
    """

    return query_str


from google.cloud import bigquery
import pandas as pd


def list_tables(dataset_id, project_id='basedosdados', pattern=None):
    """
    List tables inside a dataset_id

    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id. 
        project_id (:obj:`str`, optional): In case you want to use to 
            query another project, by default 'basedosdados'
        pattern (:obj:`str`, optional): Regular expression to look 
            within `dataset_id`.

    Returns: 
        pd.DataFrame: List of `table_id` within the `dataset_id`
    """
    client = bigquery.Client(project=project_id)
    
    tables_list = client.list_tables(dataset_id)

    result = pd.DataFrame(dict(
        table_id=[table.table_id for table in tables_list]
    ))
    # filter tables using Regex
    if pattern: 
        result = result.loc[result['table_id'].str.contains(pattern)]
    
    return result

def list_datasets(project_id='basedosdados', pattern=None):
    """
    List `dataset_id`'s inside a `project_id`

    Args:
        pattern (:obj:`str`, optional): Regular expression to look 
            within `dataset_id`.

    Returns: 
        pd.DataFrame: List of `table_id` within the `dataset_id`
    """

    client = bigquery.Client(project=project_id)
    datasets_list = client.list_datasets()

    result = pd.DataFrame(dict(
        dataset_id=[table.dataset_id for table in datasets_list]
    ))

    # filter tables using Regex
    if pattern: 
        result = result.loc[result['dataset_id'].str.contains(pattern)]
    
    return result

def metadata(dataset_id, table_id, project_id='basedosdados'):
    """
    Display column types and descriptions from `table_id`
    
    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        table_id (:obj:`str`): Table id available in project_id.dataset_id.
        project_id (:obj:`str`, optional): In case you want to use to 
            query another project, by default 'basedosdados'

    Returns: 
        pd.DataFrame: Column types and descriptions
    """
    client = bigquery.Client(project=project_id)
    table_name = f'{project_id}.{dataset_id}.{table_id}'

    table = client.get_table(table_name)
    description = [
        (col.name, col.field_type, col.description) for col in table.schema
            ]
    # TODO: full table description
    # check df.metadata

    return pd.DataFrame(description, columns=['columns', 'type','description'])

def cost(query, price=0.02):
    """
    Estimate the cost of a query in BigQuery considering each Gb costs.
    Up to now, it only works with real tables (not on Views or External)

    Args:
        query (:obj:`str`): Valid SQL Standard Query.
        price (:obj:`float`, optional): Price in US dollars per GB consumed. 

    Returns:
        float: Cost of query in US dollars
    """
    client = bigquery.Client()

    job_config = bigquery.QueryJobConfig()
    job_config.dry_run = True
    job_config.use_query_cache = False
    
    job = client.query(query, location='US', job_config=job_config)
    mb = _pretty_format(job.total_bytes_processed, 'bytes')

    # US$ for each GB processed.
    dollar = (job.total_bytes_processed / 1024/1024/1024) * price

    print(f'Processing {mb} MB, which costs US$ {dollar:,.2f}')
        
    return dollar

def info(
    dataset_id, 
    table_id, 
    project_id='basedosdados', 
    pretty=True, 
    extract_from_view=False):
    """
    Display metadata about the specified `table_id` 

    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        table_id (:obj:`str`): Table id available in project_id.dataset_id.
        project_id (:obj:`str`, optional): In case you want to use to query 
            another project, by default 'basedosdados'
        pretty (:obj:`bool`, optional): Whether to return values as 
            raw data (for calculation purposes) or pretty formatted.
        extract_from_view: Run a query to count number of rows and bytes used
            in a view. Be aware that this will incurr additional cost.
    Returns: 
        pd.DataFrame: Metadata information about the table

    """

    client = bigquery.Client()

    table_name = f'{project_id}.{dataset_id}.{table_id}'

    table = client.get_table(table_name)
    
    nrows = table.num_rows
    nbytes = table.num_bytes
    
    # when tables are views, using table.num_rows returns 0
    if extract_from_view:
        job = client.query(
            f'SELECT COUNT(*) FROM {table_name}', 
            location='US'
        )
        nrows = job.to_dataframe().loc[0, 'f0_']
        nbytes = job.total_bytes_processed 

    # For bytes, though, if you are working with public data from bq
    # the job will not account for any byte, so I'll get the maximum 
    # between that and the table.num_bytes
    
    name = table.table_id
    dataset = table.dataset_id
    fullname = table.full_table_id
    location = table.location
    anomes_created = table.created.date()
    time_created = table.created.time()
    description = table.description

    if pretty:
        nrows = f'{nrows:,}'
        nbytes = _pretty_format(nbytes, 'bytes')
        anomes_created = str(table.created.year) + \
                '/' + str(table.created.month).zfill(2) + \
                '/' + str(table.created.day).zfill(2)
        time_created = str(table.created.hour) + \
            ':' + str(table.created.minute).zfill(2)
        
    data = pd.DataFrame({'nrows': [nrows], 
                         'size': [nbytes],
                         'table_name':[name],
                         'dataset_name':[dataset],
                         'date_created':[anomes_created],
                         'time_created':[time_created],
                         'full_table_name':[fullname],
                         'table_location':[location],
                         'description':[description]}, index=['value']).T

    return data

def _pretty_format(value, category='value'):
    """
    Converts unfriendly formats to human readable.

    Args:
        value (:obj:`int`, float): The value to be formatted.
        category (:obj:`str`, optional): The type of value.
            Currently allowed categories are: 
            - `value` for numbers to be written as 80K, 80M, 80B
            - `bytes` for numbers to be written as 1MB, 1kB, 1GB

    Returns: 
        str: formatted value
    """
    if category == 'value':
        units = ['','K','M','B']
        decimal_places = {'' : 0, 
                          'K':1,
                          'M':1,
                          'B':1
                        }
        cutoff = 1000
    elif category == 'bytes':
        units = ['B','kB','MB','GB','TB']
        decimal_places = {'B' : 3, 
                          'kB':2,
                          'MB':1,
                          'GB':1,
                          'TB':1
                            }
        cutoff = 1024
    else:

        raise NotImplementedError()

    def _human_readable_size(size, 
                            units=units, 
                            decimal_places=decimal_places):
        for unit in units:
            if size < cutoff:
                break
            size /= cutoff

        return f"{size:.{decimal_places[unit]}f}{unit}"

    return _human_readable_size(value)

def _retrieve_information_schema(dataset_id, project_id='basedosdados'):
    client = bigquery.Client()
    job = client.query(
        f'SELECT * FROM `{project_id}.{dataset_id}.INFORMATION_SCHEMA.VIEWS`'
    )
    return job.to_dataframe()





from basedosdados.dataset import Dataset
from basedosdados.storage import Storage
from basedosdados.table import Table
from basedosdados.download.metadata import metadata, list_tables, list_datasets, info
from basedosdados.download.download import read_sql, download, read_table
from basedosdados.download import stats



from google.cloud import bigquery
import pandas as pd

from basedosdados.download.base import google_client


def list_tables(dataset_id, project_id="basedosdados", pattern=None):
    """
    List tables inside a dataset_id

    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        project_id (:obj:`str`, optional): In case you want to use to
            query another project, by default 'basedosdados'
        pattern (:obj:`str`, optional): Regular expression to look
            within `dataset_id`.

    Returns:
        pd.DataFrame: List of `table_id` within the `dataset_id`
    """
    client = google_client("bigquery", project_id=project_id)

    tables_list = client.list_tables(dataset_id)

    result = pd.DataFrame(dict(table_id=[table.table_id for table in tables_list]))
    # filter tables using Regex
    if pattern:
        result = result.loc[result["table_id"].str.contains(pattern)]

    return result


def list_datasets(project_id="basedosdados", pattern=None, with_tables=False):
    """
    List `dataset_id`'s inside a `project_id`

    Args:
        pattern (:obj:`str`, optional): Regular expression to look
            within `dataset_id`.
        with_table (:obj:`bool`, optional): Whether to return the tables
            related to the dataset. It might take a while.

    Returns:
        pd.DataFrame: List of `table_id` within the `dataset_id`
    """

    client = google_client("bigquery", project_id=project_id)
    datasets_list = client.list_datasets()

    result = pd.DataFrame(
        dict(dataset_id=[table.dataset_id for table in datasets_list])
    )

    # filter tables using Regex
    if pattern:
        result = result.loc[result["dataset_id"].str.contains(pattern)]

    # load table names
    if with_tables:
        result = (
            result.groupby("dataset_id")["dataset_id"]
            .apply(lambda x: list_tables(x.values[0]))
            .reset_index()
            .drop("level_1", 1)
        )

    return result


def metadata(dataset_id, table_id, project_id="basedosdados"):
    """
    Display column types and descriptions from `table_id`

    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        table_id (:obj:`str`): Table id available in project_id.dataset_id.
        project_id (:obj:`str`, optional): In case you want to use to
            query another project, by default 'basedosdados'

    Returns:
        pd.DataFrame: Column types and descriptions
    """
    client = google_client("bigquery", project_id=project_id)
    table_name = f"{project_id}.{dataset_id}.{table_id}"

    table = client.get_table(table_name)
    description = [(col.name, col.field_type, col.description) for col in table.schema]
    # TODO: full table description
    # check df.metadata

    return pd.DataFrame(description, columns=["columns", "type", "description"])


def cost(query, price=0.02):
    """
    Estimate the cost of a query in BigQuery considering each Gb costs.
    Up to now, it only works with real tables (not on Views or External)

    Args:
        query (:obj:`str`): Valid SQL Standard Query.
        price (:obj:`float`, optional): Price in US dollars per GB consumed.

    Returns:
        float: Cost of query in US dollars
    """
    client = google_client("bigquery")

    job_config = bigquery.QueryJobConfig()
    job_config.dry_run = True
    job_config.use_query_cache = False

    job = client.query(query, location="US", job_config=job_config)
    mb = _pretty_format(job.total_bytes_processed, "bytes")

    # US$ for each GB processed.
    dollar = (job.total_bytes_processed / 1024 / 1024 / 1024) * price

    print(f"Processing {mb} MB, which costs US$ {dollar:,.2f}")

    return dollar


def info(
    dataset_id,
    table_id=None,
    project_id="basedosdados",
    pretty=True,
    extract_from_view=False,
):
    """
    Display metadata about the specified `dataset_id` and/or `table_id`.

    Use just the `dataset_id` to get info about a dataset.

    Args:
        dataset_id (:obj:`str`): Dataset id available in project_id.
        table_id (:obj:`str`, optional): Table id available in project_id.dataset_id.
        project_id (:obj:`str`, optional): In case you want to use to query
            another project, by default 'basedosdados'
        pretty (:obj:`bool`, optional): Whether to return values as
            raw data (for calculation purposes) or pretty formatted.
        extract_from_view: Run a query to count number of rows and bytes used
            in a view. Be aware that this will incurr additional cost.
    Returns:
        pd.DataFrame: Metadata information about the table

    """

    client = google_client("bigquery", project_id=project_id)

    if table_id:
        table_name = f"{project_id}.{dataset_id}.{table_id}"

        table = client.get_table(table_name)

        nrows = table.num_rows
        nbytes = table.num_bytes

        # when tables are views, using table.num_rows returns 0
        if extract_from_view:
            job = client.query(f"SELECT COUNT(*) FROM {table_name}", location="US")
            nrows = job.to_dataframe().loc[0, "f0_"]
            nbytes = job.total_bytes_processed

        # For bytes, though, if you are working with public data from bq
        # the job will not account for any byte, so I'll get the maximum
        # between that and the table.num_bytes

        name = table.table_id
        dataset = table.dataset_id
        fullname = table.full_table_id
        location = table.location
        anomes_created = table.created.date()
        time_created = table.created.time()
        description = table.description

        if pretty:
            nrows = f"{nrows:,}"
            nbytes = _pretty_format(nbytes, "bytes")
            anomes_created = (
                str(table.created.year)
                + "/"
                + str(table.created.month).zfill(2)
                + "/"
                + str(table.created.day).zfill(2)
            )
            time_created = (
                str(table.created.hour) + ":" + str(table.created.minute).zfill(2)
            )

        data = pd.DataFrame(
            {
                "nrows": [nrows],
                "size": [nbytes],
                "table_name": [name],
                "dataset_name": [dataset],
                "date_created": [anomes_created],
                "time_created": [time_created],
                "full_table_name": [fullname],
                "table_location": [location],
                "description": [description],
            },
            index=["value"],
        ).T

    else:

        dataset = client.get_dataset(dataset_id)

        anomes_created = (
            str(dataset.created.year)
            + "/"
            + str(dataset.created.month).zfill(2)
            + "/"
            + str(dataset.created.day).zfill(2)
        )

        time_created = (
            str(dataset.created.hour) + ":" + str(dataset.created.minute).zfill(2)
        )

        data = pd.DataFrame(
            {
                "dataset_name": [dataset.dataset_id],
                "date_created": [anomes_created],
                "time_created": [time_created],
                "full_dataset_name": [dataset.full_dataset_id],
                "location": [dataset.location],
                "description": [dataset.description],
            },
            index=["value"],
        ).T

    return data


def _pretty_format(value, category="value"):
    """
    Converts unfriendly formats to human readable.

    Args:
        value (:obj:`int`, float): The value to be formatted.
        category (:obj:`str`, optional): The type of value.
            Currently allowed categories are:
            - `value` for numbers to be written as 80K, 80M, 80B
            - `bytes` for numbers to be written as 1MB, 1kB, 1GB

    Returns:
        str: formatted value
    """
    if category == "value":
        units = ["", "K", "M", "B"]
        decimal_places = {"": 0, "K": 1, "M": 1, "B": 1}
        cutoff = 1000
    elif category == "bytes":
        units = ["B", "kB", "MB", "GB", "TB"]
        decimal_places = {"B": 3, "kB": 2, "MB": 1, "GB": 1, "TB": 1}
        cutoff = 1024
    else:

        raise NotImplementedError()

    def _human_readable_size(size, units=units, decimal_places=decimal_places):
        for unit in units:
            if size < cutoff:
                break
            size /= cutoff

        return f"{size:.{decimal_places[unit]}f}{unit}"

    return _human_readable_size(value)


def _retrieve_information_schema(dataset_id, project_id="basedosdados"):
    client = google_client("bigquery", project_id=project_id)
    job = client.query(
        f"SELECT * FROM `{project_id}.{dataset_id}.INFORMATION_SCHEMA.VIEWS`"
    )
    return job.to_dataframe()



from functools import lru_cache
from google.cloud import bigquery, storage
import os
from pathlib import Path
import pandas_gbq

import re
import time

from basedosdados.download.base import google_client, credentials


def read_sql(query, project_id="basedosdados"):
    """Load data from BigQuery using a query. Just a wrapper around pandas.read_gbq

    Args:
        query (str): Valid SQL Standard Query.

    Returns:
        pd.DataFrame: The query result.

    Examples:
        >>> my_query = '''
        SELECT partido,
               COUNT(*) AS qtd_deputados
        FROM `basedosdados.br_sp_alesp.deputados`
        GROUP BY partido
        '''
        >>> bd.read_sql(query=my_query)

        |    | partido      |   qtd_deputados |
        |---:|:-------------|------:|
        |  0 | PSL          |    13 |
        |  1 | PROS         |     1 |
        |  2 | PSD          |     2 |
        |  3 | SD           |     1 |
        |  4 | REPUBLICANOS |     6 |
        |  ... | ...         |     ... |
    """

    try:
        return pandas_gbq.read_gbq(
            query, credentials=credentials(), project_id=project_id
        )
    except OSError:
        raise OSError(
            "The project could not be determined.\n"
            "Set the project with `gcloud config set project <project_id>`.\n"
            "Where <project_id> is your Google Cloud Project ID that can be found "
            "here https://console.cloud.google.com/projectselector2/home/dashboard \n"
        )


def read_table(dataset_id, table_id, project_id="basedosdados", limit=None):
    """Load data from BigQuery using dataset_id and table_id.

    Args:
        dataset_id (str): Dataset id available in project_id.
        table_id (str): Table id available in project_id.dataset_id.
        project_id (str, optional): In case you want to use to query another
            project, defaults to 'basedosdados'
        limit (int, optional): Number of rows.

    Returns:
        pd.DataFrame: The query result
    """

    if (dataset_id is not None) and (table_id is not None):
        query = f"""
        SELECT * 
        FROM `{project_id}.{dataset_id}.{table_id}`"""

        if limit is not None:

            query += f" LIMIT {limit}"
    else:
        raise Exception("Both table_id and dataset_id should be filled.")

    return read_sql(query, project_id=project_id)
